/*
 * Atomic operation macros.
 *
 * This file defines the kernel atomic operations available on the TC2000
 * and GP1000.
 * Some of these are routines defined using augmentations (TC2000) or the
 * PNC (GP1000) in atomic.c (TC2000) or pnctrap.s (GP1000). Others are
 * macros built on those primitives.
 *
 * nX on the TC2000 and GP1000 supports the same set of atomic operations,
 * with the exception that there is no 'Atomic_cta32()' available
 * on the GP.
 */

#ifndef _atomic_h
#define _atomic_h yes
#ifdef KERNEL
#include "bbn_vvrmt_debug.h"
#else KERNEL
#include <sys/features.h>
#endif KERNEL

#ifdef BFLY2
/*
 * The primitive operations are based on Atomic_cta16, Atomic_cta32,
 * and xmemsupword
 */
/*
 * Clear then Add
 * Returns old contents of the cell.
 */
int Atomic_ctadd16(/*address, mask, incr*/);
int Atomic_ctadd32(/*address, mask, incr*/);


/*
 * The rest of the operations can be implemented using those two.
 */

#define Atomic_and32(ptr, mask) \
    Atomic_ctadd32((ptr), ~(mask), 0)

#define Atomic_ior32(ptr, mask) \
    Atomic_ctadd32((ptr), (mask), (mask))

#define Atomic_add32(ptr, incr) \
    Atomic_ctadd32((ptr), 0, (incr))

#define Atomic_clear32(ptr) \
    xmemsupword((ptr), 0)

/*
 * There is no 'phys_Atomic_ior16()' operation on the TC.
 */
#define phys_Atomic_ior16(ptr, mask) ERROR ERROR ERROR

#endif BFLY2

#ifdef BFLY1
/*
 * GP1000 versions. These are based on routines in pnctrap.s
 */

int Atomic_add32(/*ptr, incr*/);
int Atomic_and32(/*ptr, mask*/); 
int Atomic_ior32(/*ptr, mask*/);
int Atomic_ctadd16(/*ptr, mask, incr*/);

/*
 * No 32 bit clear then add on GP
 */
#define Atomic_ctadd32(ptr, mask, incr) ERROR ERROR ERROR

/*
 *
 */
#define Atomic_clear32(ptr) Atomic_and32(ptr, 0)

/*
 * Physical inclusive or. Used only for setting an IPI on a remote
 * processor (sinterrupt.c)
 */
int phys_Atomic_ior16(/*ptr, mask*/);

#endif BFLY1

/*
 * Operations implemented the same on both architectures
 */

#define Atomic_add16(ptr, incr) \
    Atomic_ctadd16((ptr), 0, (incr))

#define Atomic_and16(ptr, mask) \
    Atomic_ctadd16((ptr), ~(mask), 0)


#define Atomic_ior16(ptr, mask) \
    Atomic_ctadd16((ptr), (mask), (mask))

/*
 * Atomic_clear16 - fetch a value and replace it with 0
 * There is no xmem.h instruction on the 88K
 */
#define Atomic_clear16(ptr) \
    Atomic_ctadd16((ptr), -1, 0)

/*
 * Atomic_fetch[16/32] - return the current value of a location.
 * On the GP and TC this is a simple fetch, however the operation
 * is defined (and should be used) for the case of a machine which
 * does not provide hardware atomic operations.
 */
#define Atomic_fetch16(ptr) (*((short *)(ptr)))
#define Atomic_fetch32(ptr) (*((int *)(ptr)))

/*
 * Atomic_store[16/32] - store a value atomically. See Atomic_fetch
 * for justification.
 */
#define Atomic_store16(ptr,value) *((short *)(ptr)) = (value)
#define Atomic_store32(ptr,value) *((long *)(ptr)) = (value)

/* Vectorized queue operations.  These routines manipulate queues
 *   whose heads live in the vectorized variable area near the start
 *   of the kernel text segment. We make use of the F9 mapping established
 *   by pmap_bootstrap, see the comment there to understand the mapping. 
 *
 * NB: Routine-based versions with consistency checks also available for
 *	debugging.
 */
#define VECTORIZEDAREA_SIZE VM_PAGESIZE

#if BBN_VVRMT_DEBUG==0
/*
 * vvrmtptr - return a pointer to vectorized variable in F9 space
 * for the appropriate processor.
 *
 * Note that if the size of the vectorized page changes this macro must
 * change.
 */
#define vvrmtptr(ptr, node) \
    ((vm_offset_t)(((u_long)(ptr)&(VECTORIZEDAREA_SIZE-1))+\
		   ((node)*VECTORIZEDAREA_SIZE)+\
		   0xF9000000))

/*
 * vvlocptr - return a pointer to the local copy of a vectorized
 * variable.
 */
extern char start_vectorized;
#define vvlocptr(ptr) \
    ((vm_offset_t)(((u_long)(ptr)&(VECTORIZEDAREA_SIZE-1)) +\
		   (u_long)&start_vectorized))

/* Different consistency checks for routine version in atomic.c */
#define	vvlocptr_maybe(ptr)	vvlocptr(ptr)

/*
 * vvlognode - return the logical node associated with a vectorized
 * variable
 */
#define vvlognode(ptr) \
    ((((u_long)(ptr) - 0xf9000000))/VECTORIZEDAREA_SIZE)
#endif BBN_VVRMT_DEBUG
#endif _atomic_h
