/*
 ****************************************************************
 * Mach Operating System
 * Copyright (c) 1986 Carnegie-Mellon University
 *  
 * This software was developed by the Mach operating system
 * project at Carnegie-Mellon University's Department of Computer
 * Science. Software contributors as of May 1986 include Mike Accetta, 
 * Robert Baron, William Bolosky, Jonathan Chew, David Golub, 
 * Glenn Marcy, Richard Rashid, Avie Tevanian and Michael Young. 
 * 
 * Some software in these files are derived from sources other
 * than CMU.  Previous copyright and other source notices are
 * preserved below and permission to use such software is
 * dependent on licenses from those institutions.
 * 
 * Permission to use the CMU portion of this software for 
 * any non-commercial research and development purpose is
 * granted with the understanding that appropriate credit
 * will be given to CMU, the Mach project and its authors.
 * The Mach project would appreciate being notified of any
 * modifications and of redistribution of this software so that
 * bug fixes and enhancements may be distributed to users.
 *
 * All other rights are reserved to Carnegie-Mellon University.
 ****************************************************************
 */

#ifndef	_MESSAGE_
#define	_MESSAGE_	1

/*
 *	Mach IPC message and primitive function definitions.
 *
 * HISTORY
 * 10-Dec-86  Avadis Tevanian (avie) at Carnegie-Mellon University
 *	Fixed macros for detecting whether or not an error code is
 *	really an error code.
 *
 * 10-Dec-86 Mary Thompson
 *	changed RCV_ERROR_END from -207 to -208
 *
 *  3-Dec-86  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Documented use of MSG_TYPE_EMERGENCY as a bit only.
 *
 * 10-Nov-86  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Added MSG_TYPE_INVALID.
 *
 * 21-Oct-86  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Made msg_type_long_number a normal "int" rather than a "long".
 *
 *  5-Oct-86  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Added include of <sys/boolean.h>, even though fields are
 *	declared as integer bit fields.  Forced boolean values
 *	to be exported.
 *
 * 20-Sep-86  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Eliminated SEND_PORT_FULL error return... never possible.
 *
 * 19-Sep-86  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Fixed MSG_TYPE_PORT_NAME.  Added SEND_MSG_TOO_LARGE/SMALL errors.
 *
 *  6-Sep-86  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Added "lint library" declarations of the message traps.
 *
 * 16-Aug-86  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Added some more error returns from msg_send/receive; added
 *	"byte" type for the kernel.
 *
 * 23-Jul-86  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Added IPC type names.
 *
 * 20-Jun-86  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Created.
 */

#define	EXPORT_BOOLEAN

#ifdef	KERNEL
#include "../h/port.h"
#include "../h/boolean.h"
#else	KERNEL
#include <sys/port.h>
#include <sys/boolean.h>
#endif	KERNEL

/*
 *	Message data structures.
 *
 *	Messages consist of two parts: a fixed-size header, immediately
 *	followed by a variable-size array of typed data items.
 *
 */

typedef	struct {
		int		:24,
				msg_simple : 8;
		int		msg_size;
		int		msg_type;
		port_t		msg_local_port;
		port_t		msg_remote_port;
		int		msg_id;
} msg_header_t;

#define	MSG_SIZE_MAX	8192

/*
 *	Known values for the msg_type field.
 *	These are Accent holdovers, which should be purged when possible.
 *
 *	Only one bit in the msg_type field is used by the kernel.
 *	Others are available to user applications.  See <msg_type.h>
 *	for system application-assigned values.
 */

#define	MSG_TYPE_NORMAL		0
#define	MSG_TYPE_EMERGENCY	1

/*
 *	Each data item is preceded by a description of that
 *	item, including what type of data, how big it is, and
 *	how many of them are present.
 *
 *	The actual data will either follow this type
 *	descriptor ("inline") or will be specified by a pointer.
 *
 *	If the type name, size, or number is too large to be encoded
 *	in this structure, the "longform" option may be selected,
 *	and those fields must immediately follow in full integer fields.
 *
 *	For convenience, out-of-line data regions or port rights may
 *	be deallocated when the message is sent by specifying the
 *	"deallocate" field.  Beware: if the data item in question is both
 *	out-of-line and contains port rights, then both will be deallocated.
 */

typedef struct  {
	unsigned int	msg_type_name : 8,		/* What kind of data */
			msg_type_size : 8,		/* How many bits is each item */
			msg_type_number : 12,		/* How many items are there */
			msg_type_inline : 1,		/* If true, data follows; else a pointer */
			msg_type_longform : 1,		/* Name, size, number follow: see above */
			msg_type_deallocate : 1;	/* Deallocate port rights or memory */
} msg_type_t;

typedef	struct	{
	msg_type_t	msg_type_header;
	short		msg_type_long_name;
	short		msg_type_long_size;
	int		msg_type_long_number;
} msg_type_long_t;

/*
 *	Known values for the msg_type_name field.
 *
 *	The only types known to the Mach kernel are
 *	the port types, and those types used in the
 *	kernel RPC interface.
 */

#define	MSG_TYPE_UNSTRUCTURED	0
#define	MSG_TYPE_BIT		0
#define	MSG_TYPE_BOOLEAN	0
#define	MSG_TYPE_INTEGER_16	1
#define	MSG_TYPE_INTEGER_32	2
#define	MSG_TYPE_PORT_OWNERSHIP	3
#define	MSG_TYPE_PORT_RECEIVE	4
#define	MSG_TYPE_PORT_ALL	5
#define	MSG_TYPE_PORT		6
#define	MSG_TYPE_CHAR		8
#define	MSG_TYPE_BYTE		9
#define	MSG_TYPE_INTEGER_8	9
#define	MSG_TYPE_REAL		10
#define	MSG_TYPE_STRING		12
#define	MSG_TYPE_INVALID	13		/* Used by the kernel to mark invalid data */
#define	MSG_TYPE_LAST		14		/* Last assigned */

#define	MSG_TYPE_PORT_NAME	2		/* XXX Should be a new type */

/*
 *	Is a given item a port type?
 */
#define	MSG_TYPE_PORT_ANY(x)	( ((x) >= MSG_TYPE_PORT_OWNERSHIP) && ((x) <= MSG_TYPE_PORT) )

/*
 *	Options to IPC primitives.
 *
 *	These can be combined by or'ing; the combination RPC call
 *	uses both SEND_ and RCV_ options at once.
 */

typedef	int		msg_option_t;

#define	MSG_OPTION_NONE	0x0000		/* Terminate only when message op works */

#define	SEND_TIMEOUT	0x0001		/* Terminate on timeout elapsed */
#define	SEND_NOTIFY	0x0002		/* Terminate with reply message if need be */
	
#define	RCV_TIMEOUT	0x0100		/* Terminate on timeout elapsed */
#define	RCV_NO_SENDERS	0x0200		/* Terminate if I'm the only sender left */


/*
 *	Returns from IPC primitives.
 *
 *	Values are separate in order to allow RPC users to
 *	distinguish which operation failed; for successful completion,
 *	this doesn't matter.
 */

typedef	int		msg_return_t;

#define	SEND_SUCCESS		0

#define	SEND_ERRORS_START	-100
#define	SEND_INVALID_MEMORY	-101		/* Message structure or out-of-line data invalid */
#define	SEND_INVALID_PORT	-102		/* Reference to inacessible port */
#define	SEND_TIMED_OUT		-103		/* Terminated due to timeout */
#define	SEND_WILL_NOTIFY	-105		/* Last message accepted until notification sent */
#define	SEND_NOTIFY_IN_PROGRESS	-106		/* Already awaiting a notification */
#define	SEND_KERNEL_REFUSED	-107		/* Message to the kernel refused */
#define	SEND_INTERRUPTED	-108		/* Software interrupt during send */
#define	SEND_MSG_TOO_LARGE	-109		/* Message specified was too large */
#define	SEND_MSG_TOO_SMALL	-110		/* Data specified exceeds message size */
#define	SEND_ERRORS_END		-111

#define	msg_return_send(x)	((x) < SEND_ERRORS_START && (x) > SEND_ERRORS_END)

#define	RCV_SUCCESS		0

#define	RCV_ERRORS_START	-200
#define	RCV_INVALID_MEMORY	-201
#define	RCV_INVALID_PORT	-202
#define	RCV_TIMED_OUT		-203
#define	RCV_TOO_LARGE		-204		/* Message structure too small to hold data */
#define	RCV_NOT_ENOUGH_MEMORY	-205		/* Unable to find address space to put out-of-line data */
#define	RCV_ONLY_SENDER		-206		/* Receiver/owner are only remaining senders */
#define	RCV_INTERRUPTED		-207
#define	RCV_ERRORS_END		-208

#define	msg_return_rcv(x)	((x) < RCV_ERRORS_START && (x) > RCV_ERRORS_END)

#define	RPC_SUCCESS		0

/*
 *	The IPC primitive functions themselves
 */

msg_return_t	msg_send(
#if	LINTLIBRARY
			header, option, timeout)
	msg_header_t	*header;
	msg_option_t	option;
	int		timeout;
	{ return(SEND_SUCCESS); }
#else	LINTLIBRARY
		);
#endif	LINTLIBRARY

msg_return_t	msg_receive(
#if	LINTLIBRARY
			header, option, timeout)
	msg_header_t	*header;
	msg_option_t	option;
	int		timeout;
	{ return(RCV_SUCCESS); }
#else	LINTLIBRARY
		);
#endif	LINTLIBRARY

msg_return_t	msg_rpc(
#if	LINTLIBRARY
			header, option, rcv_size,
			send_timeout, rcv_timeout)
	msg_header_t	*header;	/* in/out */
	msg_option_t	option;
	int		rcv_size;
	int		send_timeout;
	int		rcv_timeout;
	{ return(RPC_SUCCESS); }
#else	LINTLIBRARY
		);
#endif	LINTLIBRARY


#endif	_MESSAGE_
