/* Copyright 1987 Bolt Beranek and Newman Inc. */
/* All Rights Reserved */
/*
 * xylogics.h
 *
 * This file contains the definitions for the Xylogics 450 Disk Controller
 * registers and parameter blocks.
 */

/*
 * This is the standard I/O parameter block used for every command to the
 * controller.
 *
 * HISTORICAL NOTE:
 * Two forms of this structure are given:  one for use in byte
 * swapped mode and one for use in non-byte-swapped mode.  Which one you
 * should use at any instant depends upon whether the Multibus adaptor is
 * set for byte swaps or not.  Remember that the Multibus byte order is the
 * opposite of the one for the 68000.
 *
 * The assumed naming convention here is that you will use struct iopb if you
 * do not have the Multibus adaptor set for byte swaps and iopb_with_swaps
 * if you do have it set for byte swaps.
 *
 * UPDATE (7/86):
 * The BMA has been found to do Multibus -> PNC word/byte address mappings
 * improperly (after a year and a half of use).  The IOPB is always read
 * using byte-mode DMA.  Now, with byte-mode DMA, the BMA performs the
 * same mapping independent of whether MB_SWAP or MB_NOSWAP is used.
 */
 
typedef struct {
	unsigned char comm;	/* command byte */
	unsigned char imode;	/* interrupt mode / function modification */
	unsigned char stat1;	/* status byte 1 */
	unsigned char ccode;	/* completion code (status byte 2) */
	unsigned char throttle;	/* DMA throttling value */
	unsigned char dtype_usel; /* drive type and unit select byte */
#define DRIVE_TYPE_SHIFT 6
	unsigned char head;	/* head byte */
	unsigned char sector;	/* sector byte */
	unsigned char cyl_lo;	/* low byte of cylinder number */
	unsigned char cyl_hi;	/* high byte of cylinder number */
	unsigned char scount_lo; /* low byte of sector count */
	unsigned char scount_hi; /* high byte of sector count */
	unsigned char daddr_lo;	/* data address low byte */
	unsigned char daddr_hi;	/* data address high byte */
	unsigned char drel_lo;	/* data relocation value low byte */
	unsigned char drel_hi;	/* data relocation value high byte */
	unsigned char hoff;	/* head offset register */
#define ENABLE_DRIVE_TYPE_ZERO 0x40
	unsigned char subcode;	/* operation subcode (only used for format */
	unsigned char nextiopb_lo; /* low byte of next iopb address */
	unsigned char nextiopb_hi; /* high byte of next iopb address */

	/*
	 * The next two bytes are reversed in order;
	 * this is not an error, but simply the way the controller writes
	 * them.  It may have something to do with the fact that the
	 * pattern is bit reversed.
	 */

	unsigned short ecc_pat;	  /* bit reversed ECC pattern */
	unsigned char ecc_off_lo; /* ECC offset address, low byte */
	unsigned char ecc_off_hi; /* ECC offset address, high byte */
} IOPB_NO_SWAPS;

#ifdef neverdef
/*
 * HISTORICAL:
 * See "Programming the Butterfly/Multibus Adapter" section 2.1 about byte
 * swapping and the BMA.
 */
typedef struct {
	unsigned char imode;	/* interrupt mode / function modification */
	unsigned char comm;	/* command byte */
	unsigned char ccode;	/* completion code (status byte 2) */
	unsigned char stat1;	/* status byte 1 */
	unsigned char dtype_usel; /* drive type and unit select byte */
	unsigned char throttle;	/* DMA throttling value */
	unsigned char sector;	/* sector byte */
	unsigned char head;	/* head byte */
	unsigned short cyl;	/* cylinder number */
	unsigned short scount;	/* sector count */
	unsigned short daddr;	/* data address */
	unsigned short drel;	/* data relocation value */
	unsigned char subcode;	/* operation subcode (only used for format */
	unsigned char hoff;	/* head offset register */
	unsigned short nextiopb; /* next iopb address */

	/*
	 * The next two bytes are reversed in order;
	 * this is not an error, but simply the way the controller writes
	 * them.  It may have something to do with the fact that the
	 * pattern is bit reversed.
	 */

	unsigned char ecc_pat_hi; /* bit reversed ECC pattern high byte*/
	unsigned char ecc_pat_lo; /* bit reversed ECC pattern low byte */
	unsigned short ecc_off;	/* ECC offset address */
} IOPB_WITH_SWAPS;
#endif

/*
 * Define XYLOGICS 450 control registers in Multibus I/O space.
 * As explained in section 1.6 of the manual, these registers are
 * byte addressable only.
 */
typedef struct {
	unsigned char	iopb_rrlb;	/* relocation register low byte */
	unsigned char	iopb_rrhb;	/* relocation register high byte */
	unsigned char	iopb_arlb;	/* address register low byte */
	unsigned char	iopb_arhb;	/* address register high byte */
	unsigned char	csr;		/* control and status register */
	unsigned char	crur;		/* controller reset/update register */
} XYLOGICS;

/* Default location of control registers in Multibus I/O space */
#define DEFAULT_IOADDR	0xffe4ee40	/* no swap */

/* Control and status register definitions */

#define CSR_GBSY	0x80	/* (R/W) go/busy bit */
#define CSR_ERR		0x40	/* (R/W) general error, error reset */
#define CSR_DERR	0x20	/* (R) double error */
#define CSR_IPND	0x10	/* (R/W) interrupt pending */
#define CSR_ADRM	0x08	/* (R) addressing mode */
#define CSR_AREQ	0x04	/* (R/W) attention request */
#define CSR_AACK	0x02	/* (R) attention acknowledge */
#define CSR_DRDY	0x01	/* (R) drive ready */

/* Command byte definitions (IOPB byte 0) */

#define COMM_AUD	0x80	/* auto update */
#define COMM_RELO	0x40	/* relocation */
#define COMM_CHEN	0x20	/* chaining enable */
#define COMM_IEN	0x10	/* interrupt enable */
#define COMM_COM	0x0F	/* command value mask */
# define COMM_COM_NOP			0x00
# define COMM_COM_WRITE			0x01
# define COMM_COM_READ			0x02
# define COMM_COM_WRITE_TRACK_HEADERS 	0x03
# define COMM_COM_READ_TRACK_HEADERS 	0x04
# define COMM_COM_SEEK			0x05
# define COMM_COM_DRIVE_RESET		0x06
# define COMM_COM_WRITE_FORMAT		0x07
# define COMM_COM_READ_HEADER_DATA_ECC	0x08
# define COMM_COM_READ_DRIVE_STATUS	0x09
# define COMM_COM_WRITE_HEADER_DATA_ECC	0x0A
# define COMM_COM_SET_DRIVE_SIZE	0x0B
# define COMM_COM_SELF_TEST		0x0C
# define COMM_COM_DMA_TEST		0x0D
# define COMM_COM_MAINT_BUFFER_LOAD	0x0E
# define COMM_COM_MAINT_BUFFER_DUMP	0x0F

/*
 * Interrupt mode / function modification (IMODE) bit definitions
 * (IOPB byte 1)
 */

#define IMODE_IEI	0x40	/* Interrupt on each IOPB */
#define IMODE_IERR	0x20	/* Interrupt on error */
#define IMODE_HDP	0x10	/* Hold Dual Port Driver bit */
#define IMODE_ASR	0x08	/* Auto Seek Retry */
#define IMODE_EEF	0x04	/* Enable Extended Function */
# define IMODE_ECC_STOP		0x00	/* Report ECC pattern and offset,
					 * Stop a chained transfer,
					 * Report an ECC error (1E or 06),
					 * Lose at least one revolution
					 */
# define IMODE_ECC_IGNORE	0x01	/* Does not correct or flag an error,
					 * Continues an IOPB chain on soft
					 * error,
					 * Does not lose a revolution.
					 */
# define IMODE_ECC_CORRECT	0x02	/* Corrects error, if possible,
					 * Updates IOPB with ECC error status,
					 * Continues an IOPB chain on soft
					 * error,
					 * Loses one revolution.
					 */
# define IMODE_ECC_FLAG		0x03	/* Does not correct an error,
					 * Flags an ECC error (code 06)
					 * Continues an IOPB chain,
					 * Does not lose a revolution.
					 */

/* Status byte 1 (IOPB byte 2) bit definitions */

#define STAT1_ERRS	0x80	/*  error summary bit, set if any error */

#define STAT1_CTYP_MASK	0x1C	/* Mask for controller type bits */
	/* example: ((x & STAT1_CTYP_MASK) == STAT1_CTYP_450) */
# define STAT1_CTYP_440		0x00	/* 440 controller */
# define STAT1_CTYP_450		0x04	/* 450 controller */
# define STAT1_CTYP_472		0x08	/* 472 controller */

#define STAT1_DONE	0x01	/* operation complete */

/* Status byte 2 (IOPB byte 3) completion code values */

#define STAT2_SUCCESS		0	/* successful completion of operation */
#define STAT2_INT_PEND		1	/* interrupt pending */
#define STAT2_BUSY		3	/* busy conflict */
#define STAT2_TIMEOUT		4	/* operation timeout */
#define STAT2_HNF		5	/* header not found */
#define STAT2_HARD_ECC		6	/* hard ecc error */
#define STAT2_ILL_CYL		7	/* illegal cylinder address */
#define STAT2_ILL_SECT		0xa	/* illegal sector address */
#define STAT2_RUNT_SECT		0xd	/* last sector too small */
#define STAT2_NOMEM		0xe	/* slave ACK error: non-exist memory */
#define STAT2_HDR_ERR		0x12	/* cylinder and head/header error */
#define STAT2_SEEK_RETRY	0x13	/* seek retry required */
#define STAT2_WRITE_PROT	0x14	/* write-protect error */
#define STAT2_DRIVE_NOT_READY	0x16
#define STAT2_SCZ		0x17	/* sector count zero */
#define STAT2_DRIVE_FAULTED	0x18
#define STAT2_ILL_SECT_SIZE	0x19
#define STAT2_SELF_TEST_A	0x1a
#define STAT2_SELF_TEST_B	0x1b
#define STAT2_SELF_TEST_C	0x1c
#define STAT2_SOFT_ECC		0x1e
#define STAT2_SOFT_ECC_FIXED	0x1f
#define STAT2_ILL_HEAD		0x20
#define STAT2_DSE		0x21	/* disk sequencer error */
#define STAT2_SEEK_ERROR	0x25

/*
 * Throttle values to determine DMA burst size (in bytes)
 */
#define THRO_2		0
#define THRO_4		1
#define THRO_8		2
#define THRO_16		3
#define THRO_32		4
#define THRO_64		5
#define THRO_128	6
#define THRO_256	7

/* Copyright 1987 Bolt Beranek and Newman Inc. */
/* All Rights Reserved */

typedef IOPB_NO_SWAPS	IOPB;
/* since have to byte/word ordering manipulations for IOPB: */
#define hibyte_of_word(w)	(((w) >> 8) & 0xff)
#define lobyte_of_word(w)	((w) & 0xff)
#define hiword_of_long(l)	(((l) >> 16) & 0xffff)
#define loword_of_long(l)	((l) & 0xffff)
#define construct_word(h,l)	(((h) << 8) + (l))
#define construct_long(h,l)	(((h) << 16) + (l))


typedef char BOOLEAN;

#define DISKS_PER_CONTROLLER 4

#define BYTES_PER_SECTOR	512

typedef int OPERATION_STATE;
#define OP_SUCCEEDED	((OPERATION_STATE) 1)
#define OP_RETRIED	((OPERATION_STATE) 2)
#define OP_FAILED	((OPERATION_STATE) 3)

/*
 * Default disk parameters used so that we can read the configuration
 * information in from disk.  We correct the drive's configuration after
 * we know what it really is.  Defaults are for Fujitsu Eagle 1K sectors,
 * 1 spare per track.
 */
#define DEF_SECTORS	 47
#define DEF_TRACKS	 20
#define DEF_CYLINDERS	842
#define DEF_SPARE	  1

typedef struct disk_size {
	unsigned short	nsectors;	/* sectors per track */
	unsigned short	ntracks;	/* tracks per cylinder */
	unsigned short	cyl;		/* cylinders per disk */
	unsigned int	nbytes;		/* bytes per sector */
} DISK_CONFIG;
